/* We define some stuff in this file before the headers because
   the headers #define symbols used in these functions */

/* Verbose and somewhat useful malloc/free tracing */
/* turn it on by hand-munging the obvious spot in headers.h */

#ifdef TRACE_MALLOC
void *MYmalloc(size_t size) {
  void *result = malloc(size);
  fprintf(stderr, "Allocated %6d at %x\n", size, result);
  return result;
}
char *MYcalloc(size_t nmemb, size_t size) {
  char *result = calloc(nmemb,size);
  fprintf(stderr, "Allocated %6d at %x\n", nmemb*size, result);
  return result;
}
void MYfree(void *ptr){
  fprintf(stderr, "freeing %x\n", ptr);
  free(ptr);
}
#endif

/* I guess we need to see the standard headers here, even though
   they're massive overkill (but then, massive header overkill is
   traditional in C programming, isn't it?
*/
#include "headers.h"
/* we use POSIX regex in this file */
#include <regex.h>
/* and errno */
#include <errno.h>


/* code support for SS error "handling" */
void spit_status(int stat, char const *call) {
  if(stat < 0) {
    fprintf(stderr, "ERROR: code %d result of %s\n", stat, call);
    exit(-1);
  }
}

/* more code for fatal error reporting */
/* this should really use varargs, but that's a royal pain */
void fatal_error(char const *s1, char const *s2)
{
  fprintf(stderr, "FATAL ERROR: %s %s\n", s1, s2);
  exit(-1);
}


/* We use calloc to allocate string buffers because many of the HDF-EOS
   string-buffer-smashing routines do *NOT* null-terminate the strings
   they smash into the buffers.  I can see why not (FORTRAN compatibility),
   but it does make for counter-intuitive C processing...
*/
char *newstring(unsigned int len) {
  char *result = calloc(len, 1);
  if(result == NULL)
    spit_status(-1, "Ran out of memory allocating a little bitty string!");
  return result;
}

/* a little string copier - surely this is a library function somewhere? */
char *strclone(char const *in)
{
  char *result = (char *) malloc(strlen(in) + 1);
  strcpy(result, in);
  return result;
}

/* a little wrapper around POSIX regular expression matching */
/* returns 1 if strings match, 0 if they don't, calls fatal_error
   if the regex is malformed
*/

int regex_matches(char const *str, char const *pattern)
{
  /* As another convenience, matching is by default anchored -
     the match string implicitly begins with ^.
  */
  regex_t pbs;
  int errcode;
  char *anchorpattern = newstring(strlen(pattern)+2);
  int result = 1;

  sprintf(anchorpattern, "^%s", pattern);

  if((errcode = regcomp(&pbs, anchorpattern, REG_EXTENDED|REG_NOSUB))) {
    size_t ebsize = regerror(errcode, &pbs, NULL, 0);
    char *errbuf = newstring(ebsize + 1);

    regerror(errcode, &pbs, errbuf, ebsize);
    fatal_error(errbuf, pattern);
  }
  else {
    if(regexec(&pbs, str, 0, NULL, 0) == REG_NOMATCH) {
      result = 0;
    }
    free(anchorpattern);
  }
  return result;
}

/* convert the string to an (unsigned) long - set bogus flag if string isn't
   a valid base 10 integer
*/
long careful_atol(char const *a, H5T_sign_t signp, int *bogus)
{
  char *endptr = NULL;
  long result = (signp ? strtol(a, &endptr, 10)
		       : (long)strtoul(a, &endptr, 10));

  *bogus = 
     (signp ? ((result == LONG_MIN || result == LONG_MAX) && errno == ERANGE)
             : ((unsigned long) result == ULONG_MAX && errno == ERANGE));

  return result;
}

/* next_name and end_name take pointers to a name buffer
   (one of those comma-seperated atrocities the HDF inquiry functions return).
   end_name checks to see if the current name is terminated by \0 or ',', and
   clobbers any ',' with \0.
   next_name just returns the pointer moved forward to the next field.
   This means you can write loops of the form:

   nameptr = namebuf;
   for(i=0; i< numnames; i++) {
     end_name(nameptr);
     ... do stuff with nameptr
     nameptr = next_name(nameptr);
   }

   avoiding the extra malloc/freeing needed with their breakout utility.
*/

void end_name(char *namebuf)
{
  while(*namebuf != 0 && *namebuf != ',') namebuf++;
  if(*namebuf == ',') *namebuf = 0;
}

char *next_name (char *namebuf)
{
  char *result = namebuf;
  if(namebuf == NULL || *namebuf == 0) return result;
  while(*result != 0 && *result != ',') result++;
  return ++result;
}

/* this returns the "index" of the name in the list,
   (zero-based, as God and Dennis Richie intended),
   or -1 if name is not in the list
*/
int name_in_list(char const *name, char *namelist, int namecount)
{
  int i;

  for(i=0; i< namecount; i++) {
    end_name(namelist);
    if(!strcmp(name, namelist)) return i;
    namelist = next_name(namelist);
  }
  return -1;
}
/* A simple stack-based location name remembering system, to handle
   the hierarchical locations inside an HDF-EOS file.  Would you believe
   this would actually be easier in C++?
*/

typedef struct namestack {
  struct namestack *next;
  char *type;
  char *name;
  int printed;
} namestack;

namestack *current = NULL, *old = NULL;

namestack *new_ns(void);
namestack *new_ns(void)
{
  namestack *result;
  if (old == NULL)
    result = (namestack *) malloc(sizeof(namestack));
  else {
    result = old;
    old = old->next;
  }
  result->printed = 0;
  return result;
}

void recycle_ns(namestack *foo);
void recycle_ns(namestack *foo)
{
  foo->next = old;
  old = foo;
}

void push_name(char const *type, char const *name)
{
  namestack *new = new_ns();
  new->type = strclone(type);
  new->name = strclone(name);
  new->next = current;
  current = new;
  
}

void pop_name()
{
  if(current != NULL) {
    namestack *now = current;
    free(now->name);
    free(now->type);
    current = current->next;
    recycle_ns(now);
  }
}

/* tabcount counts the number of names skipped coming up the stack */
int tabcount;

void print_name_temp(namestack *foo, FILE *outstream);
void print_name_temp(namestack *foo, FILE *outstream)
{
  if(foo != NULL) {
    print_name_temp(foo->next, outstream);
    /*    if(!(foo->printed)) {
      if(tabcount > 0)
	putc('\n', outstream);
      for( ; tabcount > 0; tabcount--)
      fprintf(outstream, "    "); */
      fprintf(outstream, " %s %s", foo->type, foo->name);
      /*      foo->printed = 1;
    }
    else tabcount++;*/
  }
}

void print_name(FILE *outstream)
{
  tabcount = 0;
  putc('\n', outstream);
  print_name_temp(current, outstream);
}

char const *current_name()
{
  if (current != NULL)
    return current->name;
  else return "";
}

char const *current_type()
{
  if(current != NULL)
    return current->type;
  else return "";
}

/* Error counter and error code support */

int error_count = 0;

void bump_errors() { error_count++; }

int print_final_message() {
  if(error_count) {
  fprintf(stderr, "\nErrors:\t\t%d\n",
	  error_count);
  }
  return -(error_count);
}
